import pandas as pd

def convertVC(vc: pd.Series, colName: str):
    return vc.reset_index().rename( 
        columns={"index": "value", 
        colName: "count"}).to_dict('records')

def convertDescribe(statistics: pd.Series):
    s = statistics.to_dict()

    return {
        "min": s["min"],
        "q25": s["25%"],
        "q50": s["50%"],
        "q75": s["75%"],
        "max": s["max"],
        "mean": s["mean"],
    }

def convertBinned(numVC: pd.Series, true_min):
    """
    numVC has interval index from binned value counts. Replace far left low 
    with true min because pandas cuts below min
    """
    d = pd.DataFrame(
        {"low": numVC.index.left,
         "high": numVC.index.right,
         "count": numVC.values
        })
    d = d.reset_index().rename(columns={"index": "bucket"})
    d_dict = d.to_dict('records')

    if len(d_dict) > 0:
        d_dict[0]['low'] = true_min
    
    return d_dict


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Will Epperson.
# Distributed under the terms of the Modified BSD License.

"""
Visualizer module for widgets
"""

from ipywidgets import DOMWidget
from traitlets import Unicode, Dict, observe
import pandas as pd
from varname import argname
from ipylab import JupyterFrontEnd

from ._frontend import module_name, module_version
from .profile_lib import isNumeric, isTimestamp, isCategorical, getShape, \
    getColMeta, getValueCounts, getQuantBinnedData, getTempBinnedData, getTempInterval, \
    getQuantMeta, getStringMeta, getTemporalMeta
from .utils import convertVC, convertBinned

class Visualizer(DOMWidget):
    # boilerplate for ipywidgets syncing
    _model_name = Unicode('VizualizerModel').tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode('VizualizerView').tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)

    # our synced traitlet state
    dfProfile = Dict({}).tag(sync=True)
    exportedCode = Unicode('').tag(sync=True)

    # python only state
    dataframe = None
    app = JupyterFrontEnd()
    

    def __init__(self, dataframe: pd.DataFrame, *args, **kwargs):
        super(Visualizer, self).__init__(*args, **kwargs)

        if not isinstance(dataframe, pd.DataFrame):
            raise ValueError("dataframe must be a pandas DataFrame!")

        self.dataframe = dataframe
        dfName = argname('dataframe')
        self.calculateChartData(dfName)
    
    def calculateChartData(self, dfName: str):
        # get columns and check that all names are unique
        df = self.dataframe
                
        if df.columns.nunique() != len(df.columns):
            raise ValueError("Column names are not unique!")
        
        shape = getShape(self.dataframe)
        colProfiles = []

        # Get data for each column
        for cName in df.columns:
            vc = getValueCounts(df, cName, isIndex=False)
            num_unique, num_null = getColMeta(df, cName, isIndex=False)

            cd = {
                "name": cName,
                "type": str(df[cName].dtype),
                "isIndex": False,
                "summary": {
                    "cardinality": num_unique,
                    "topK": convertVC(vc, cName) # TODO this might not be right type
                },
                "nullCount": num_null,
                "example": vc.index[0]
            }

            if isNumeric(df[cName]):
                # get data
                chartData = getQuantBinnedData(df, cName, isIndex=False)
                statistics = getQuantMeta(df, cName, isIndex=False)

                # convert to JSON serializable
                chartData = convertBinned(chartData, statistics["min"])

                cd["summary"]["quantMeta"] = statistics
                cd["summary"]["histogram"] = chartData
            elif isTimestamp(df[cName]):
                # get data
                vc, true_min = getTempBinnedData(df, cName, isIndex=False)
                interval = getTempInterval(df, cName, isIndex=False)
                temporalMeta = getTemporalMeta(df, cName, isIndex=False)

                # convert to JSON serializable
                histogram = convertBinned(vc, true_min)

                cd["summary"]["histogram"] = histogram
                cd["summary"]["timeInterval"] = interval
                cd["summary"]["temporalMeta"] = temporalMeta

            elif isCategorical(df[cName]):
                stringMeta = getStringMeta(df, cName)

                cd["summary"]["stringMeta"] = stringMeta
            
            colProfiles.append(cd)
    
        # This should be JSON serializable
        profile = {
            "profile": colProfiles,
            "shape": shape,
            "dfName": dfName,
            "lastUpdatedTime": 0,
            "isPinned": False,
            "warnings": []
        }
        # save profile to trailet to sync with frontend
        self.dfProfile = profile

    @observe('exportedCode')
    def _observe_exported_code(self, change):
        """
        Called when the exportedCode traitlet is changed, we add new code cell on a change
        """
        self.addNewCell(change['new'])
        self.exportedCode = ''
    
    def addNewCell(self, codeText):
        if codeText == '':
            return
        self.app.commands.execute('notebook:insert-cell-below')
        self.app.commands.execute('notebook:replace-selection', {'text': codeText})



"""
Utility functions for calculating metadata from pandas dataframes
These are called from PythonExecutor.ts in the frontend
"""

import pandas as pd
from pandas.api.types import is_string_dtype, is_numeric_dtype, is_bool_dtype, is_categorical_dtype, is_datetime64_dtype
from .utils import convertDescribe

####### type checks #######
def isNumeric(colData: pd.Series):
    return is_numeric_dtype(colData)

def isTimestamp(colData: pd.Series):
    return is_datetime64_dtype(colData)

def isCategorical(colData: pd.Series):
    return is_categorical_dtype(colData) or is_string_dtype(colData)

def isBoolean(colData: pd.Series):
    return is_bool_dtype(colData)

######## Analysis functions #######
def getColumns(dfName: pd.DataFrame):
    typeDF = dfName.dtypes.reset_index().rename(columns={"index": "colName", 0: "type"})
    typeDF["isIndex"] = False
    if dfName.index.name:
        name = dfName.index.name
    else:
        name = ""
    indexDF = pd.DataFrame({"colName": [name], "type": [dfName.index.dtype]})
    indexDF["isIndex"] = True
    typeDF = pd.concat([indexDF, typeDF])
    return typeDF

def getShape(dfName: pd.DataFrame):
    return dfName.shape

def getColMeta(dfName: pd.DataFrame, colName: str, isIndex=False):
    if isIndex:
        colData = dfName.index.to_series()
    else:
        colData = dfName[colName]
    num_unique = colData.nunique()
    num_null = colData.isna().sum()

    return num_unique, num_null

def getValueCounts(dfName: pd.DataFrame, colName: str, n=10, isIndex=False):
    if isIndex:
        colData = dfName.index.to_series()
    else:
        colData = dfName[colName]
    vc = colData.value_counts().iloc[:n]
    return vc

def getQuantBinnedData(dfName: pd.DataFrame, colName: str, n=20, isIndex=False):
    if isIndex:
        colData = dfName.index.to_series()
    else:
        colData = dfName[colName]
    vc = colData.value_counts(bins=min(n, colData.nunique()), sort=False)
    return vc

def getTempBinnedData(dfName: pd.DataFrame, colName: str, n=200, isIndex=False):
    if isIndex:
        colData = dfName.index.to_series()
    else:
        colData = dfName[colName]
    vc = (colData.astype("int64")//1e9).value_counts(bins=min(n, colData.nunique()), sort=False)
    true_min = (colData.astype("int64")//1e9).min()
    return vc, true_min

def getTempInterval(dfName: pd.DataFrame, colName: str, isIndex=False):
    if isIndex:
        colData = dfName.index.to_series()
    else:
        colData = dfName[colName]
    timerange = colData.max() - colData.min()
    return {"months": 0, "days": timerange.days, "micros": 0}

def getStringMeta(dfName: pd.DataFrame, colName: str, isIndex=False):
    if isIndex:
        lengths = dfName.index.to_series().str.len()
    else:
        lengths = dfName[colName].str.len()

    return {
        "minLength": lengths.min(),
        "maxLength": lengths.max(),
        "meanLength": lengths.mean(),
    }

def getQuantMeta(dfName: pd.DataFrame, colName: str, isIndex=False):
    if isIndex:
        colData = dfName.index.to_series()
    else:
        colData = dfName[colName]
    
    describe = colData.describe()
    sd = describe.loc['std']
    mean = describe.loc['mean']
    q3 = describe.loc['75%']
    q1 = describe.loc['25%']

    # get num outliers > 3 std away from mean
    normalized = (colData - mean) / sd
    sd_num_outliers = sum( abs(normalized) > 3)

    # get iqr outliers that are 1.5 * iqr away from q1 or q3
    iqr = q3 - q1
    lower = q1 - 1.5 * iqr
    upper = q3 + 1.5 * iqr
    iqr_num_outliers = sum((colData < lower) | (colData > upper))

    # get sortedness
    if colData.is_monotonic_increasing:
        sortedness = "ascending"
    elif colData.is_monotonic_decreasing:
        sortedness = "descending"
    else:
        sortedness = "noSort"
    
    n_zero = sum(colData == 0)
    n_negative = sum(colData < 0)
    n_positive = sum(colData > 0)

    # make serializable

    statistics = convertDescribe(describe)
    statistics["sd_outlier"] = sd_num_outliers
    statistics["iqr_outlier"] = iqr_num_outliers
    statistics["sortedness"] = sortedness
    statistics["n_zero"] = n_zero
    statistics["n_positive"] = n_positive
    statistics["n_negative"] = n_negative

    return statistics

def getTemporalMeta(dfName:pd.DataFrame, colName:str, isIndex=False):
    if isIndex:
        colData = dfName.index
    else:
        colData = dfName[colName]

    if colData.is_monotonic_increasing:
        result = "ascending"
    elif colData.is_monotonic_decreasing:
        result = "descending"
    else:
        result = "noSort"
    return {"sortedness": result}
